<?php

namespace Drupal\workers_ai_provider\Plugin\AiProvider;

use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\ai\Attribute\AiProvider;
use Drupal\ai\Base\AiProviderClientBase;
use Drupal\ai\Enum\AiModelCapability;
use Drupal\ai\Exception\AiRateLimitException;
use Drupal\ai\OperationType\Chat\ChatInput;
use Drupal\ai\OperationType\Chat\ChatInterface;
use Drupal\ai\OperationType\Chat\ChatMessage;
use Drupal\ai\OperationType\Chat\ChatOutput;
use Drupal\ai\Traits\OperationType\ChatTrait;
use OpenAI\Client;
use Symfony\Component\Yaml\Yaml;

/**
 * Plugin implementation of the 'workers ai' provider.
 */
#[AiProvider(
  id: 'workers_ai',
  label: new TranslatableMarkup('Workers AI'),
)]
class WorkersAiProvider extends AiProviderClientBase implements
  ContainerFactoryPluginInterface,
  ChatInterface {

  use ChatTrait;

  /**
   * The OpenAI Client for API calls.
   *
   * @var \OpenAI\Client|null
   */
  protected $client;

  /**
   * API Key.
   *
   * @var string
   */
  protected string $apiKey = '';

  /**
   * API Key.
   *
   * @var string
   */
  protected string $cloudfareBaseURL = '';

  /**
   * API Key.
   *
   * @var string
   */
  protected string $cloudfareAccountId = '';

  /**
   * {@inheritdoc}
   */
  public function getConfiguredModels(?string $operation_type = NULL, array $capabilities = []): array {
    // No vision support.
    $cloudfareBaseURL = $this->getConfig()->get('cloudfare_base_url');
    $cloudfareAccountId = $this->getConfig()->get('cloudfare_account_id');
    $base_url = $cloudfareBaseURL . '/accounts/' . $cloudfareAccountId;
    if (in_array(AiModelCapability::ChatWithImageVision, $capabilities)) {
      return [];
    }
    try {
      $client = \Drupal::httpClient();
      $response = $client->request('GET', $base_url. "/ai/models/search", [
        'headers' => [
          'Authorization' => "Bearer pKum0EeqkErDqfIBwqDThOAkxRloUGHH5Wg7GMXI",
          'Content-Type' => 'application/json',
        ],
      ]);
    
      $body = $response->getBody()->getContents();
      $response_data = json_decode($body, TRUE);

    } catch (\Exception $e) {
      \Drupal::logger('workers_ai')->error('API request failed: @message', ['@message' => $e->getMessage()]);
      return [];
    }

    $models = [];
    $model_names =[];
    if (isset($response_data['result'])) {
      foreach ($response_data['result'] as $model) {
        $models[$model['name']] = $model['name'];
      }
    }

    if (in_array(AiModelCapability::ChatJsonOutput, $capabilities)) {
      foreach ($models as $model_id => $model_name) {
        if (strpos($model_id, 'llama') === FALSE) {
          unset($models[$model_id]);
        }
      }
    }
    return $models;
  }

  /**
   * {@inheritdoc}
   */
  public function isUsable(?string $operation_type = NULL, array $capabilities = []): bool {
    // No vision support.
    if (in_array(AiModelCapability::ChatWithImageVision, $capabilities)) {
      return FALSE;
    }
    // If its not configured, it is not usable.
    if (!$this->getConfig()->get('api_key')) {
      return FALSE;
    }
    else{
      $this->apiKey = $this->getConfig()->get('api_key');
    }
    if (!$this->getConfig()->get('cloudfare_base_url')) {
      return FALSE;
    }
    else{
      $this->cloudfareBaseURL = $this->getConfig()->get('cloudfare_base_url');
    }
    if (!$this->getConfig()->get('cloudfare_account_id')) {
      return FALSE;
    }
    else{
      $this->cloudfareAccountId = $this->getConfig()->get('cloudfare_account_id');
    }
    // If its one of the bundles that Workers AI supports its usable.
    if ($operation_type) {
      return in_array($operation_type, $this->getSupportedOperationTypes());
    }
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedOperationTypes(): array {
    return [
      'chat',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getConfig(): ImmutableConfig {
    return $this->configFactory->get('workers_ai_provider.settings');
  }

  /**
   * {@inheritdoc}
   */
  public function getApiDefinition(): array {
    // Load the configuration.
    $definition = Yaml::parseFile($this->moduleHandler->getModule('workers_ai_provider')->getPath() . '/definitions/api_defaults.yml');
    return $definition;
  }

  /**
   * {@inheritdoc}
   */
  public function getModelSettings(string $model_id, array $generalConfig = []): array {
    return $generalConfig;
  }

  /**
   * {@inheritdoc}
   */
  public function setAuthentication(mixed $authentication): void {
    // Set the new API key and reset the client.
    $this->apiKey = $authentication;
    $this->client = NULL;
  }

  /**
   * Gets the raw client.
   *
   * This is the client for inference.
   *
   * @return \OpenAI\Client
   *   The OpenAI client.
   */
  public function getClient(): Client {
    $this->loadClient();
    return $this->client;
  }

  /**
   * Loads the Workers AI Client with authentication if not initialized.
   */
  protected function loadClient(): void {
    if (!$this->client) {
      if (!$this->apiKey) {
        $this->setAuthentication($this->loadApiKey());
      }
      $cloudfareBaseURL = $this->getConfig()->get('cloudfare_base_url');
      $cloudfareAccountId = $this->getConfig()->get('cloudfare_account_id'); 
      $final_base_url = $cloudfareBaseURL .'/accounts/'. $cloudfareAccountId;
      $this->client = \OpenAI::factory()
        ->withApiKey($this->apiKey)
        ->withBaseUri($final_base_url)
        ->withHttpClient($this->httpClient)
        ->make();       
    }
  }

  /**
   * {@inheritdoc}
   */
  public function chat(array|string|ChatInput $input, string $model_id, array $tags = []): ChatOutput {
    $cloudfareBaseURL = $this->getConfig()->get('cloudfare_base_url');
    $cloudfareAccountId = $this->getConfig()->get('cloudfare_account_id');
    $base_url = $cloudfareBaseURL . '/accounts/' . $cloudfareAccountId;
    $chat_input = [];
  
    // Prepare chat messages.
    if ($input instanceof ChatInput) {
      if ($this->chatSystemRole) {
        $chat_input[] = [
          'role' => 'system',
          'content' => $this->chatSystemRole,
        ];
      }
      foreach ($input->getMessages() as $message) {
        $chat_input[] = [
          'role' => 'user',
          'content' => $message->getText(),
        ];
      }
    }
    // Build API request payload.
    $payload = [
      'messages' => $chat_input,
    ];
  
    try {
      $client = \Drupal::httpClient();
      $response = $client->post(
        $base_url."/ai/run/{$model_id}",
        [
          'headers' => [
            'Authorization' => 'Bearer pKum0EeqkErDqfIBwqDThOAkxRloUGHH5Wg7GMXI',
            'Content-Type' => 'application/json',
          ],
          'json' => $payload,
        ]
      );
  
      $data = json_decode($response->getBody()->getContents(), TRUE);
      $message_text = $data['result']['response'] ?? 'No response';  
      $message = new ChatMessage('assistant', $message_text);
      return new ChatOutput($message, $data, []);
    }
    catch (\Exception $e) {
      if (strpos($e->getMessage(), 'Rate limit') !== FALSE) {
        throw new AiRateLimitException($e->getMessage());
      }
      throw $e;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getSetupData(): array {
    return [
      'key_config_name' => 'api_key',
      'default_models' => [
        'chat' => '@cf/meta/llama-3.3-70b-instruct-fp8-fast',
      ],
    ];
  }

}
